import os
from typing import Union, List


class InputFile:
    def __init__(
        self,
        file: Union[str, bytes, None] = None,
        file_id: str | None = None,
        is_image: bool = True,
        as_document: bool = False,
    ):
        """
        Represents a file to be sent to Mattermost.

        :param file: Path to the file, bytes of the file, or None.
        :param file_id: Existing file ID in Mattermost (if already uploaded).
        :param is_image: Whether the file is an image.
        :param as_document: Whether to send the image as a document.
        """
        if not file and not file_id:
            raise ValueError("Either 'file' or 'file_id' must be provided.")

        self.file = file
        self.file_id = file_id
        self.is_image = is_image
        self.as_document = as_document

        if isinstance(file, str) and not os.path.isfile(file):
            raise ValueError(
                f"The file path '{file}' does not exist or is not a file."
            )

    def is_uploaded(self) -> bool:
        """Check if the file is already uploaded (identified by file_id)."""
        return self.file_id is not None

    def is_file_path(self) -> bool:
        """Check if the file is a valid file path."""
        return isinstance(self.file, str) and os.path.isfile(self.file)


class InputFileGroup:
    def __init__(self, files: List[InputFile]):
        """
        Represents a group of InputFile objects to be sent as attachments.

        :param files: List of InputFile objects.
        """
        if not files:
            raise ValueError(
                "InputFileGroup must contain at least one InputFile."
            )

        self.files = files

    def get_uploaded_file_ids(self) -> List[str]:
        """Retrieve file IDs of already uploaded files."""
        return [file.file_id for file in self.files if file.is_uploaded()]

    def get_unuploaded_files(self) -> List[InputFile]:
        """Retrieve files that need to be uploaded."""
        return [file for file in self.files if not file.is_uploaded()]



import asyncio
from logging import Logger
from typing import Any, Callable, Coroutine, Dict, List, Union

from .driver import MattermostDriver
from .events import EventType
from .plugin import Plugin
from .settings import Settings
from .aiologger import setup_logger
from .state.fsm_context import FSMContext

EventHandler = Callable[
    [Dict[str, Any], FSMContext], Coroutine[Any, Any, None]
]


class Bot:
    def __init__(
        self,
        plugins: List[Plugin],
        settings: Settings,
        logger: Logger | None = None,
    ):
        """Инициализация бота с подключенными плагинами и настройками."""
        self.plugins = plugins
        self.settings = settings
        self.logger = logger or setup_logger()
        self.driver = MattermostDriver(
            settings.full_api_url, settings.BOT_TOKEN, logger=self.logger
        )
        self.handlers: Dict[EventType, List[EventHandler]] = (
            self._register_plugins()
        )
        self.fsm_contexts: Dict[str, FSMContext] = (
            {}
        )  # FSMContext для каждого пользователя

    def _register_plugins(self) -> Dict[EventType, List[EventHandler]]:
        """Регистрирует плагины и передает им драйвер."""
        handlers: Dict[EventType, List[EventHandler]] = {}
        for plugin in self.plugins:
            plugin.setup_plugin(self.driver, self.logger)
            for event_type, funcs in plugin.get_handlers().items():
                handlers.setdefault(event_type, []).extend(funcs)
        return handlers

    async def _initialize(self) -> None:
        await self.driver.initialize()

    def _get_fsm_context(self, user_id: str) -> FSMContext:
        """Получает или создает FSMContext для пользователя."""
        if user_id not in self.fsm_contexts:
            self.fsm_contexts[user_id] = FSMContext()
        return self.fsm_contexts[user_id]

    async def handle_event(self, event: Union[Dict[str, Any], str]):
        """Обрабатывает событие и вызывает соответствующие хэндлеры."""
        try:
            if isinstance(event, str):
                import json
                try:
                    event = json.loads(event)
                except json.JSONDecodeError:
                    self.logger.error("Ошибка: событие не удалось декодировать из строки.")
                    return
            if not isinstance(event, dict):
                self.logger.error("Ошибка: событие имеет некорректный формат.")
                return
            event["handled"] = False
            event_type = event.get('event')
            matching_handlers = self.handlers.get(event_type, [])
            if EventType.ANY in self.handlers:
                matching_handlers.extend(self.handlers[EventType.ANY])
            post_data = event.get("data", {}).get("post")
            if isinstance(post_data, str):
                import json
                try:
                    post_data = json.loads(post_data)
                except json.JSONDecodeError:
                    self.logger.error("Ошибка: не удалось декодировать поле 'post'.")
                    post_data = {}
            if not isinstance(post_data, dict):
                post_data = {}
            user_id = post_data.get("user_id", "default")
            fsm_context = self._get_fsm_context(user_id)
            for handler in matching_handlers:
                if event.get("handled"):
                    break
                await handler(event, fsm_context)
            if not event.get("handled"):
                self.logger.warning(f"Событие {event_type} осталось необработанным.")
        except Exception as e:
            self.logger.exception(f"Ошибка при обработке события:\n{e}")

    async def _async_run(self):
        """Асинхронный метод запуска бота."""
        await self._initialize()

        ws_url = f"{self.settings.full_api_url}/websocket"
        self.logger.info(f"Бот подключается к WebSocket: {ws_url}")

        while True:
            try:
                await self.driver.connect_websocket(ws_url, self.handle_event)
            except Exception as e:
                self.logger.error(
                    f"Ошибка WebSocket: {e}. Переподключение через 5 секунд..."
                )
                await asyncio.sleep(5)

    def run(self):
        """Синхронный метод запуска бота."""
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            loop.run_until_complete(self._async_run())
        finally:
            loop.close()

    async def _async_stop(self):
        """Асинхронная остановка бота."""
        await self.driver.close()

    def stop(self):
        """Синхронный метод для остановки бота."""
        asyncio.run(self._async_stop())



from typing import Iterator, Type


class State:
    """
    Represents a single state in a finite-state machine.
    """

    def __init__(
        self, state: str | None = None, group_name: str | None = None
    ) -> None:
        self._state = state
        self._group_name = group_name
        self._group: Type["StatesGroup"] | None = None

    @property
    def group(self) -> Type["StatesGroup"]:
        if not self._group:
            raise RuntimeError(f"State '{self._state}' is not part of any group.")
        return self._group

    @property
    def state(self) -> str:
        group = self._group_name or (
            self._group.__full_group_name__ if self._group else "@"
        )
        return f"{group}:{self._state}"

    def set_parent(self, group: Type["StatesGroup"]) -> None:
        if not issubclass(group, StatesGroup):
            raise ValueError("Parent group must be a subclass of StatesGroup")
        self._group = group

    def __set_name__(self, owner: Type["StatesGroup"], name: str) -> None:
        if self._state is None:
            self._state = name
        self.set_parent(owner)

    def __str__(self) -> str:
        return f"<State '{self.state}'>"

    def __repr__(self) -> str:
        return self.__str__()


class StatesGroupMeta(type):
    """
    Metaclass for StatesGroup to manage states and hierarchy.
    """

    def __new__(mcs, name, bases, namespace, **kwargs):
        cls = super().__new__(mcs, name, bases, namespace)

        cls.__states__ = {
            key: value
            for key, value in namespace.items()
            if isinstance(value, State)
        }
        for state_name, state in cls.__states__.items():
            state._group_name = name

        return cls

    @property
    def __full_group_name__(cls) -> str:
        return cls.__name__

    def __iter__(cls) -> Iterator[State]:
        return iter(cls.__states__.values())


class StatesGroup(metaclass=StatesGroupMeta):
    """
    Represents a group of related states in a finite-state machine.
    """



from enum import Enum, auto


class FSMStrategy(Enum):
    """
    Defines storage strategies for finite-state machines.
    """

    USER_IN_CHAT = auto()
    """State is stored for each user in a specific chat."""
    CHAT = auto()
    """State is stored globally for a chat."""
    GLOBAL_USER = auto()
    """State is stored globally for a user."""


def apply_strategy(
    strategy: FSMStrategy, chat_id: int, user_id: int
) -> tuple[int, int]:
    """
    Apply the given strategy to determine state storage keys.

    :param strategy: The strategy to apply.
    :param chat_id: The ID of the chat.
    :param user_id: The ID of the user.
    :return: Tuple of chat_id and user_id based on strategy.
    """
    if strategy == FSMStrategy.CHAT:
        return chat_id, chat_id
    if strategy == FSMStrategy.GLOBAL_USER:
        return user_id, user_id
    return chat_id, user_id



from typing import Any, Dict

from aiomatter.state.base import State


class FSMContext:
    """
    Context for managing state and data for a specific user or entity.
    """

    def __init__(self) -> None:
        self._state: State | None = None
        self._data: Dict[str, Any] = {}

    def set_state(self, state: State | None) -> None:
        self._state = state

    def get_state(self) -> State | None:
        return self._state

    def set_data(self, key: str, value: Any) -> None:
        self._data[key] = value

    def get_data(self, key: str, default: Any = None) -> Any:
        return self._data.get(key, default)

    def get_all_data(self) -> Dict[str, Any]:
        return self._data

    def reset_state(self) -> None:
        self._state = None

    def reset_data(self) -> None:
        self._data.clear()



import json
from functools import wraps
from typing import Callable, Dict, Any, Coroutine, Type
from inspect import signature
from aiomatter.events import EventType
from aiomatter.state.base import State, StatesGroup
from aiomatter.state.fsm_context import FSMContext
from aiomatter.schemas.post_events import (
    PostEvent,
    EditEvent,
    DeleteEvent,
)

from aiomatter.aiologger import logger


EVENT_MAPPING = {
    EventType.POSTED: PostEvent,
    EventType.POST_EDITED: EditEvent,
    EventType.POST_DELETED: DeleteEvent,
    EventType.ANY: lambda event: event,
}


def listen(
    event_type: EventType,
    required_state: State | None = None,
    required_group: Type[StatesGroup] | None = None,
    ignore_bots: bool = True,
):
    """
    Декоратор для регистрации обработчиков событий
    с поддержкой фильтрации по состоянию и группе состояний.

    :param event_type: Тип события, на которое реагирует обработчик.
    :param required_state: Требуемое состояние пользователя для вызова.
    :param required_group: Требуемая группа состояний для вызова обработчика.
    :param ignore_bots: Игнорировать сообщения от ботов.
    """

    def decorator(func: Callable[..., Coroutine[Any, Any, None]]):
        @wraps(func)
        async def wrapper(
            plugin,
            event: Dict[str, Any],
            fsm_context: FSMContext | None = None,
        ):
            if event.get("event") != event_type:
                return

            post_data = event.get("data", {}).get("post")
            if isinstance(post_data, str):
                try:
                    post_data = json.loads(post_data)
                    event["data"]["post"] = post_data
                except json.JSONDecodeError:
                    logger.error("Failed to decode 'post' field.")
                    return

            sender = post_data.get("sender_name", None)
            props = post_data.get("props", {})
            from_bot = props.get("from_bot", "false")

            if (from_bot == "true" or sender == "System") and ignore_bots:
                return

            current_state = fsm_context.get_state() if fsm_context else None
            if required_group:
                if not current_state:
                    return
                if current_state.group != required_group:
                    return
            if required_state and current_state != required_state:
                return

            event_class = EVENT_MAPPING.get(event_type)
            if event_class and callable(event_class):
                try:
                    event = event_class(**event)
                except Exception as e:
                    logger.error(f"Failed to map event to class {event_class}: {e}")
                    return
            func_params = list(signature(func).parameters.keys())
            if "fsm_context" in func_params:
                await func(plugin, event, fsm_context)
            else:
                await func(plugin, event)
        wrapper.event_type = event_type
        return wrapper

    return decorator



import json
from logging import Logger

import aiohttp
import websockets

from aiomatter.input.input_file import InputFile, InputFileGroup
from aiomatter.schemas.post_events import Post
from aiomatter.schemas.user import MattermostUser


class AsyncDriver:
    def __init__(self, base_url, token, logger: Logger):
        self.base_url = base_url
        self.token = token
        self.logger = logger
        self.session = None

    async def initialize(self):
        """Асинхронная инициализация сессии."""
        self.session = aiohttp.ClientSession(
            headers={"Authorization": f"Bearer {self.token}"}
        )

    async def send_request(
        self, endpoint, method='GET', data=None, files=None
    ):
        """Отправляет запрос к REST API."""
        url = f"{self.base_url}{endpoint}"
        headers = {"Authorization": f"Bearer {self.token}"}
        async with self.session.request(
            method, url, json=data, data=files, headers=headers
        ) as response:
            response.raise_for_status()
            return await response.json()

    async def connect_websocket(self, ws_url: str, on_message):
        """Устанавливает WebSocket-соединение и слушает события."""
        headers = {'Authorization': f'Bearer {self.token}'}
        try:
            async with websockets.connect(
                ws_url, additional_headers=headers
            ) as ws:
                self.logger.info("Успешно подключено к WebSocket.")
                async for message in ws:
                    try:
                        self.logger.debug(f"Получено сообщение: {message}")
                        await on_message(json.loads(message))
                    except Exception as e:
                        self.logger.error(
                            f"Ошибка при обработке сообщения: {e}"
                        )
        except websockets.exceptions.ConnectionClosed as e:
            self.logger.error(f"Соединение закрыто: {e}")
            raise e
        except Exception as e:
            self.logger.error(f"Ошибка WebSocket: {e}")
            raise e

    async def close(self):
        """Закрывает сессию для освобождения ресурсов."""
        await self.session.close()


class MattermostDriver(AsyncDriver):
    """Драйвер взаимодействия с Mattermost API на основе WebSocket."""

    async def send_message(self, channel_id, message, root_id=None):
        """Отправляет сообщение в указанный канал."""
        data = {"channel_id": channel_id, "message": message}

        if root_id is not None:
            data["root_id"] = root_id

        response = await self.send_request("/posts", method='POST', data=data)

        return Post(**response)

    async def edit_message(self, post_id, new_message):
        """Редактирует существующее сообщение по его ID."""
        data = {"message": new_message}
        response = await self.send_request(
            f"/posts/{post_id}/patch", method='PUT', data=data
        )

        return Post(**response)

    async def delete_message(self, post_id):
        """Удаляет сообщение по его ID."""
        await self.send_request(f"/posts/{post_id}", method='DELETE')

    async def _create_dm_channel(self, user_id):
        """Создает канал для личного сообщения с пользователем."""
        bot_info = await self.get_bot_info()
        data = [bot_info.id, user_id]
        response = await self.send_request(
            "/channels/direct", method='POST', data=data
        )
        return response["id"]

    async def direct_message(self, user_id, message):
        """Отправляет личное сообщение пользователю."""
        channel_id = await self._create_dm_channel(user_id)
        return await self.send_message(channel_id, message)

    async def get_user_by_name(self, username: str) -> MattermostUser:
        """Получает информацию о пользователе по его имени."""
        username = username.replace('@', '')
        response = await self.send_request(
            f"/users/username/{username}", method='GET'
        )
        return MattermostUser(**response)

    async def get_user_by_id(self, user_id) -> MattermostUser:
        """Получает информацию о пользователе по его ID."""
        response = await self.send_request(f"/users/{user_id}", method='GET')
        return MattermostUser(**response)

    async def get_bot_info(self) -> MattermostUser:
        """Возвращает информацию о боте."""
        response = await self.send_request(
            '/users/me',
            method='GET',
        )
        return MattermostUser(**response)

    async def send_attachment(
        self, channel_id: str, input_file: InputFile, caption: str = ""
    ):
        """
        Send an attachment to a Mattermost channel.

        :param channel_id: The ID of the channel to send the message to.
        :param input_file: An InputFile object representing the file to send.
        :param message: Optional message to accompany the file.
        """
        if input_file.is_uploaded():
            data = {
                "channel_id": channel_id,
                "message": caption,
                "file_ids": [input_file.file_id],
            }
            return await self.send_request("/posts", method="POST", data=data)
        files = (
            {"files": open(input_file.file, "rb")}
            if input_file.is_file_path()
            else {"files": input_file.file}
        )
        form_data = aiohttp.FormData()
        form_data.add_field("channel_id", channel_id)
        form_data.add_field("files", files["files"])

        async with self.session.post(
            f"{self.base_url}/files",
            data=form_data,
            headers={"Authorization": f"Bearer {self.token}"},
        ) as upload_response:
            if upload_response.status not in [200, 201]:
                raise ValueError(
                    f"File upload failed: {upload_response.status}"
                )
            upload_result = await upload_response.json()
        file_id = upload_result.get("file_infos", [{}])[0].get("id")
        if not file_id:
            raise ValueError("Failed to retrieve file ID after upload.")
        data = {
            "channel_id": channel_id,
            "message": caption,
            "file_ids": [file_id],
        }
        return await self.send_request("/posts", method="POST", data=data)

    async def send_attachments_group(
        self,
        channel_id: str,
        input_file_group: InputFileGroup,
        caption: str = "",
    ):
        """
        Send a group of attachments to a Mattermost channel.

        :param channel_id: The ID of the channel to send the message to.
        :param input_file_group: An InputFileGroup object containing the files to send.
        :param message: Optional message to accompany the files.
        """
        file_ids = input_file_group.get_uploaded_file_ids()

        for input_file in input_file_group.get_unuploaded_files():
            if input_file.is_file_path():
                with open(input_file.file, "rb") as file_data:
                    form_data = aiohttp.FormData()
                    form_data.add_field("channel_id", channel_id)
                    form_data.add_field("files", file_data)

                    async with self.session.post(
                        f"{self.base_url}/files",
                        data=form_data,
                        headers={"Authorization": f"Bearer {self.token}"},
                    ) as upload_response:
                        if upload_response.status not in [200, 201]:
                            raise ValueError(
                                f"File upload failed: {upload_response.status}"
                            )
                        upload_result = await upload_response.json()

                file_id = upload_result.get("file_infos", [{}])[0].get("id")
                if not file_id:
                    raise ValueError(
                        "Failed to retrieve file ID after upload."
                    )
                file_ids.append(file_id)
            elif isinstance(input_file.file, bytes):
                form_data = aiohttp.FormData()
                form_data.add_field("channel_id", channel_id)
                form_data.add_field("files", input_file.file)

                async with self.session.post(
                    f"{self.base_url}/files",
                    data=form_data,
                    headers={"Authorization": f"Bearer {self.token}"},
                ) as upload_response:
                    if upload_response.status not in [200, 201]:
                        raise ValueError(
                            f"File upload failed: {upload_response.status}"
                        )
                    upload_result = await upload_response.json()

                file_id = upload_result.get("file_infos", [{}])[0].get("id")
                if not file_id:
                    raise ValueError(
                        "Failed to retrieve file ID after upload."
                    )
                file_ids.append(file_id)
            else:
                raise ValueError("Unsupported file type for upload.")

        data = {
            "channel_id": channel_id,
            "message": caption,
            "file_ids": file_ids,
        }
        return await self.send_request("/posts", method="POST", data=data)



from typing import Any, Awaitable, Callable, Dict

from aiomatter.state.fsm_context import FSMContext


class StateMiddleware:
    """
    Middleware for injecting FSMContext into handlers.
    """

    def __init__(self) -> None:
        self._contexts: Dict[str, FSMContext] = {}

    async def __call__(
        self,
        handler: Callable[..., Awaitable[Any]],
        event: Any,
        data: Dict[str, Any],
    ) -> Any:
        user_id = str(event.get("user_id", "default"))
        context = self._contexts.setdefault(user_id, FSMContext())

        data["fsm_context"] = context
        return await handler(event, data)
